@@ -1,27 +1,35 @@
-"""Messaging library for Python"""
-from __future__ import absolute_import
+"""Messaging library for Python."""
+from __future__ import absolute_import, unicode_literals
 
 import os
+import re
 import sys
 
-from collections import namedtuple
-from types import ModuleType
+if sys.version_info < (2, 7):  # pragma: no cover
+    raise Exception('Kombu 4.0 requires Python versions 2.7 or later.')
 
-version_info_t = namedtuple(
-    'version_info_t', ('major', 'minor', 'micro', 'releaselevel', 'serial'),
-)
+from collections import namedtuple  # noqa
 
-VERSION = version_info_t(3, 0, 32, '', '')
-__version__ = '{0.major}.{0.minor}.{0.micro}{0.releaselevel}'.format(VERSION)
+__version__ = '4.0.0'
 __author__ = 'Ask Solem'
 __contact__ = 'ask@celeryproject.org'
-__homepage__ = 'http://kombu.readthedocs.org'
+__homepage__ = 'https://kombu.readthedocs.io'
 __docformat__ = 'restructuredtext en'
 
 # -eof meta-
 
-if sys.version_info < (2, 6):  # pragma: no cover
-    raise Exception('Kombu 3.1 requires Python versions 2.6 or later.')
+version_info_t = namedtuple('version_info_t', (
+    'major', 'minor', 'micro', 'releaselevel', 'serial',
+))
+
+# bumpversion can only search for {current_version}
+# so we have to parse the version here.
+_temp = re.match(
+    r'(\d+)\.(\d+).(\d+)(.+)?', __version__).groups()
+VERSION = version_info = version_info_t(
+    int(_temp[0]), int(_temp[1]), int(_temp[2]), _temp[3] or '', '')
+del(_temp)
+del(re)
 
 STATICA_HACK = True
 globals()['kcah_acitats'[::-1].upper()] = False
@@ -31,6 +39,7 @@ if STATICA_HACK:  # pragma: no cover
     # they contain.
     from kombu.connection import Connection, BrokerConnection   # noqa
     from kombu.entity import Exchange, Queue, binding           # noqa
+    from kombu.message import Message                           # noqa
     from kombu.messaging import Consumer, Producer              # noqa
     from kombu.pools import connections, producers              # noqa
     from kombu.utils.url import parse_url                       # noqa
@@ -42,16 +51,20 @@ if STATICA_HACK:  # pragma: no cover
 
 # Lazy loading.
 # - See werkzeug/__init__.py for the rationale behind this.
+from types import ModuleType  # noqa
 
 all_by_module = {
-    'kombu.connection':    ['Connection', 'BrokerConnection'],
-    'kombu.entity':        ['Exchange', 'Queue', 'binding'],
-    'kombu.messaging':     ['Consumer', 'Producer'],
-    'kombu.pools':         ['connections', 'producers'],
-    'kombu.utils.url':     ['parse_url'],
-    'kombu.common':        ['eventloop', 'uuid'],
-    'kombu.serialization': ['enable_insecure_serializers',
-                            'disable_insecure_serializers'],
+    'kombu.connection': ['Connection', 'BrokerConnection'],
+    'kombu.entity': ['Exchange', 'Queue', 'binding'],
+    'kombu.message': ['Message'],
+    'kombu.messaging': ['Consumer', 'Producer'],
+    'kombu.pools': ['connections', 'producers'],
+    'kombu.utils.url': ['parse_url'],
+    'kombu.common': ['eventloop', 'uuid'],
+    'kombu.serialization': [
+        'enable_insecure_serializers',
+        'disable_insecure_serializers',
+    ],
 }
 
 object_origins = {}
@@ -61,6 +74,7 @@ for module, items in all_by_module.items():
 
 
 class module(ModuleType):
+    """Customized Python module."""
 
     def __getattr__(self, name):
         if name in object_origins:
@@ -100,7 +114,11 @@ new_module.__dict__.update({
     '__docformat__': __docformat__,
     '__package__': package,
     'version_info_t': version_info_t,
-    'VERSION': VERSION})
+    'version_info': version_info,
+    'VERSION': VERSION,
+    'absolute_import': absolute_import,
+    'unicode_literals': unicode_literals,
+})
 
 if os.environ.get('KOMBU_LOG_DEBUG'):  # pragma: no cover
     os.environ.update(KOMBU_LOG_CHANNEL='1', KOMBU_LOG_CONNECTION='1')
